/**
 * 
 */
package org.jboss.ws.core.soap;

import static org.jboss.ws.core.soap.ThrottlingConst.CONTENT_LENGTH;
import static org.jboss.ws.core.soap.ThrottlingConst.CONTENT_RANGE;
import static org.jboss.ws.core.soap.ThrottlingConst.DISCARD_ERROR_MSG;
import static org.jboss.ws.core.soap.ThrottlingConst.RESPONSE_LIMIT;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.xml.soap.SOAPFault;
import javax.xml.ws.soap.SOAPFaultException;

import org.jboss.logging.Logger;
import org.jboss.ws.core.jaxws.SOAPFaultHelperJAXWS;

/**
 * @author amahmood
 * 
 */
public class SOAPThrottler
{
	private static Logger log = Logger.getLogger(SOAPThrottler.class);

	// The length of the response body in 8-bit bytes
	// Content-Range Where in a full body message this partial message belongs
	// Content-Range: bytes 21010-47021/47022

	/***
	 * 
	 * @param metadata
	 * @param inputStream
	 */
	public InputStream throttleRespose(Map metadata, InputStream inputStream)
	{
		InputStream out = null;

		int responseLimit = getResponseLimit(metadata);

		if (responseLimit > 0)
		{
			// copy value from request header to response header
			List<String> l2 = new ArrayList<String>();
			l2.add(Long.toString(responseLimit));
			metadata.put(RESPONSE_LIMIT, l2);

			int contentRange = getContentRange(metadata);
			int contentLength = getContentLength(metadata);

			if (log.isDebugEnabled())
			{
				StringBuffer sb = new StringBuffer();
				sb.append("responseLimit=").append(responseLimit).append(", ").append("contentRange=").append(
						contentRange).append(", ").append("contentLength=").append(contentLength);

				log.debug(sb.toString());
			}
			if (contentLength == -1 && contentRange == -1)
			{
				// calculate lenght of input stream;
				if (isOverLimitStream(inputStream, responseLimit))
					out = createFaultStream(createFaultMsg());
			}

			if ((contentRange > responseLimit) || (contentLength > responseLimit))
			{
				out = createFaultStream(createFaultMsg());

			}
		}

		return out;
	}

	/**
	 * 
	 * @param soapFault
	 * @return
	 */
	private InputStream createFaultStream(SOAPMessageImpl soapFault)
	{
		InputStream out = null;
		try
		{
			ByteArrayOutputStream soapOutput = new ByteArrayOutputStream();
			// soapFault.writeElement(soapOutput);
			soapFault.writeTo(soapOutput);

			out = new ByteArrayInputStream(soapOutput.toByteArray());
		}
		catch (Exception e)
		{
			log.error(e.getCause(), e);
		}

		return out;
	}

	/**
	 * 
	 * @return
	 */
	private SOAPMessageImpl createFaultMsg()
	{
		SOAPMessageImpl out = null;
		try
		{
			// NOTE there is a difference in SOAP1.1 and SOAP1.2 fault.
			// Currently it does not matter for client since it verifys
			// throttling from the header action.
			// Following fault if for debug purpose and human readable only.

			SOAPFault soapFault = javax.xml.soap.SOAPFactory.newInstance().createFault(DISCARD_ERROR_MSG,
					org.jboss.ws.Constants.SOAP11_FAULT_CODE_CLIENT);
			soapFault.normalize();
			SOAPFaultException fe = SOAPFaultHelperJAXWS.getSOAPFaultException(soapFault);
			out = SOAPFaultHelperJAXWS.exceptionToFaultMessage(fe);
		}
		catch (Exception e)
		{
			log.error(e.getCause(), e);
		}

		return out;
	}

	/***
	 * 
	 */
	private void trimResponse()
	{
		// TBD
	}

	/***
	 * 
	 * @param metadata
	 * @return
	 */
	public static int getResponseLimit(Map metadata)
	{
		int out = -1;

		if (metadata != null && metadata.size() > 0)
		{
			Object header = metadata.get("HEADER");

			if ((header != null) && (header instanceof Map))
			{
				Object value = ((Map) header).get(RESPONSE_LIMIT);

				if ((value != null) && (value instanceof String))
				{
					try
					{
						out = Integer.parseInt((String) value);
					}
					catch (NumberFormatException e)
					{
						log.error(e.getCause(), e);
					}
				}
				else
					log.warn("Unable to read Response-Limit from SOAP response header. Response-Limit is not defined.");
			}
			else
				log.warn("Unable to read Response-Limit from SOAP response header. Header is Null.");
		}
		else
			log.warn("Unable to read Response-Limit. Metadata is null. ");

		return out;
	}

	/***
	 * 
	 * @param metadata
	 * @return
	 */
	public static int getContentLength(Map metadata)
	{
		int out = -1;

		if (metadata != null && metadata.size() > 0)
		{
			Object list = metadata.get(CONTENT_LENGTH);

			if ((list != null) && (list instanceof List))
			{
				Object value = ((List) list).get(0);

				if ((value != null) && (value instanceof String))
				{
					try
					{
						out = Integer.parseInt((String) value);
					}
					catch (NumberFormatException e)
					{
						log.error(e.getCause(), e);
					}
				}
				else
					log
							.warn("Unable to read Content-Length from SOAP response header. Unknown data type for Content-Length.");
			}
			else
				log.warn("Unable to read Content-Length from SOAP response header. Content-Length is not defined.");
		}
		else
			log.warn("Unable to read Content-Lenght. Metadata is null. ");

		return out;
	}

	/***
	 * 
	 * @param metadata
	 * @return
	 */
	public static int getContentRange(Map metadata)
	{

		int out = -1;

		// 21010-47021/47022

		if (metadata != null && metadata.size() > 0)
		{
			Object value = metadata.get(CONTENT_RANGE);

			if ((value != null) && (value instanceof String))
			{
				String contentRageTokens[] = ((String) value).split("/");

				if (contentRageTokens != null && contentRageTokens.length > 0)
				{
					try
					{
						out = Integer.parseInt(contentRageTokens[1]);
					}
					catch (NumberFormatException e)
					{
						log.error(e.getCause(), e);
					}
				}

				else
					log
							.debug("Unable to read Content-Range from SOAP response header. Unknown format of Content-Range.");
			}
			else
				log.debug("Unable to read Content-Range from SOAP response header. Content-Range is not defined.");
		}
		else
			log.error("Unable to read Content-Range. Metadata is null. ");
		return out;
	}

	/**
	 * @param inputStream
	 * @param limit
	 * @return
	 */
	private boolean isOverLimitStream(InputStream inputStream, int limit)
	{
		boolean out = false;

		if (inputStream.markSupported())
		{ 
			try
			{
				inputStream.mark(Integer.MAX_VALUE);

				int availableToRead = inputStream.available();

				if (availableToRead <= 0)
					availableToRead = 1024;

				byte[] b = new byte[availableToRead];

				int byteRead = -1;
				int totalByteRead = 0;

				do
				{
					byteRead = inputStream.read(b);
					totalByteRead = totalByteRead + byteRead;

					if (totalByteRead > limit)
					{
						out = true;
						break;
					}
				} while (byteRead > 0);
			}
			catch (Exception e)
			{
				log.error(e.getCause(), e);
			}
			finally
			{
				try
				{
					inputStream.reset();
				}
				catch (Exception e)
				{
					log.error(e.getCause(), e);
				}
			}
		}
		else
		{
			log.error("Could not throttle response. InputStream does not support marking.");
		}

		return out;
	}
}
